@@ -15,6 +15,7 @@ from .. import text
 
 from . import gc_functions
 from . import gc_options
+from . import phonet
 
 try:
     # LibreOffice / OpenOffice
@@ -32,14 +33,14 @@ __all__ = [ "lang", "locales", "pkg", "name", "version", "author", \
             "setOption", "setOptions", "getOptions", "getDefaultOptions", "getOptionsLabels", "resetOptions", "displayOptions", \
             "ignoreRule", "resetIgnoreRules", "reactivateRule", "listRules", "displayRules", "setWriterUnderliningStyle" ]
 
-__version__ = "1.12.2"
+__version__ = "2.0.0"
 
 
 lang = "fr"
 locales = {'fr-FR': ['fr', 'FR', ''], 'fr-BE': ['fr', 'BE', ''], 'fr-CA': ['fr', 'CA', ''], 'fr-CH': ['fr', 'CH', ''], 'fr-LU': ['fr', 'LU', ''], 'fr-BF': ['fr', 'BF', ''], 'fr-BJ': ['fr', 'BJ', ''], 'fr-CD': ['fr', 'CD', ''], 'fr-CI': ['fr', 'CI', ''], 'fr-CM': ['fr', 'CM', ''], 'fr-MA': ['fr', 'MA', ''], 'fr-ML': ['fr', 'ML', ''], 'fr-MU': ['fr', 'MU', ''], 'fr-NE': ['fr', 'NE', ''], 'fr-RE': ['fr', 'RE', ''], 'fr-SN': ['fr', 'SN', ''], 'fr-TG': ['fr', 'TG', '']}
 pkg = "grammalecte"
 name = "Grammalecte"
-version = "1.12.2"
+version = "2.0.0"
 author = "Olivier R."
 
 # Modules
@@ -70,7 +71,7 @@ def load (sContext="Python", sColorType="aRGB"):
     global _oTokenizer
     try:
         _sAppContext = sContext
-        _oSpellChecker = SpellChecker("fr", "fr-allvars.bdic", "", "")
+        _oSpellChecker = SpellChecker("fr", "fr-allvars.json", "", "")
         _oSpellChecker.activateStorage()
         _oTokenizer = _oSpellChecker.getTokenizer()
         gc_functions.load(sContext, _oSpellChecker)
@@ -235,7 +236,7 @@ class TextParser:
         self.sSentence = ""
         self.sSentence0 = ""
         self.nOffsetWithinParagraph = 0
-        self.lToken = []
+        self.lTokens = []
         self.dTokenPos = {}         # {position: token}
         self.dTags = {}             # {position: tags}
         self.dError = {}            # {position: error}
@@ -246,7 +247,7 @@ class TextParser:
         s = "===== TEXT =====\n"
         s += "sentence: " + self.sSentence0 + "\n"
         s += "now:      " + self.sSentence  + "\n"
-        for dToken in self.lToken:
+        for dToken in self.lTokens:
             s += '#{i}\t{nStart}:{nEnd}\t{sValue}\t{sType}'.format(**dToken)
             if "lMorph" in dToken:
                 s += "\t" + str(dToken["lMorph"])
@@ -267,6 +268,8 @@ class TextParser:
             self.parseText(self.sText, self.sText0, True, 0, sCountry, dOpt, bShowRuleId, bDebug, bContext)
         except:
             raise
+        self.lTokens = None
+        self.lTokens0 = None
         if bFullInfo:
             lParagraphErrors = list(self.dError.values())
             lSentences = []
@@ -279,18 +282,21 @@ class TextParser:
                     self.sSentence = sText[iStart:iEnd]
                     self.sSentence0 = self.sText0[iStart:iEnd]
                     self.nOffsetWithinParagraph = iStart
-                    self.lToken = list(_oTokenizer.genTokens(self.sSentence, True))
-                    self.dTokenPos = { dToken["nStart"]: dToken  for dToken in self.lToken  if dToken["sType"] != "INFO" }
+                    self.lTokens = list(_oTokenizer.genTokens(self.sSentence, True))
+                    self.dTokenPos = { dToken["nStart"]: dToken  for dToken in self.lTokens  if dToken["sType"] != "INFO" }
                     if bFullInfo:
-                        dSentence = { "nStart": iStart, "nEnd": iEnd, "sSentence": self.sSentence, "lToken": list(self.lToken) }
-                        for dToken in dSentence["lToken"]:
-                            if dToken["sType"] == "WORD":
-                                dToken["bValidToken"] = _oSpellChecker.isValidToken(dToken["sValue"])
-                        # the list of tokens is duplicated, to keep all tokens from being deleted when analysis
+                        self.lTokens0 = list(self.lTokens)  # the list of tokens is duplicated, to keep tokens from being deleted when analysis
                     self.parseText(self.sSentence, self.sSentence0, False, iStart, sCountry, dOpt, bShowRuleId, bDebug, bContext)
                     if bFullInfo:
-                        dSentence["lGrammarErrors"] = list(self.dSentenceError.values())
-                        lSentences.append(dSentence)
+                        for dToken in self.lTokens0:
+                            _oSpellChecker.setLabelsOnToken(dToken)
+                        lSentences.append({
+                            "nStart": iStart,
+                            "nEnd": iEnd,
+                            "sSentence": self.sSentence0,
+                            "lTokens": self.lTokens0,
+                            "lGrammarErrors": list(self.dSentenceError.values())
+                        })
                         self.dSentenceError.clear()
                 except:
                     raise
@@ -382,48 +388,38 @@ class TextParser:
     def update (self, sSentence, bDebug=False):
         "update <sSentence> and retokenize"
         self.sSentence = sSentence
-        lNewToken = list(_oTokenizer.genTokens(sSentence, True))
-        for dToken in lNewToken:
+        lNewTokens = list(_oTokenizer.genTokens(sSentence, True))
+        for dToken in lNewTokens:
             if "lMorph" in self.dTokenPos.get(dToken["nStart"], {}):
                 dToken["lMorph"] = self.dTokenPos[dToken["nStart"]]["lMorph"]
             if "aTags" in self.dTokenPos.get(dToken["nStart"], {}):
                 dToken["aTags"] = self.dTokenPos[dToken["nStart"]]["aTags"]
-        self.lToken = lNewToken
-        self.dTokenPos = { dToken["nStart"]: dToken  for dToken in self.lToken  if dToken["sType"] != "INFO" }
+        self.lTokens = lNewTokens
+        self.dTokenPos = { dToken["nStart"]: dToken  for dToken in self.lTokens  if dToken["sType"] != "INFO" }
         if bDebug:
             echo("UPDATE:")
             echo(self)
 
-    def _getNextPointers (self, dToken, dGraph, dPointer, bDebug=False):
-        "generator: return nodes where <dToken> “values” match <dNode> arcs"
-        dNode = dGraph[dPointer["iNode"]]
-        iToken1 = dPointer["iToken1"]
+    def _getMatches (self, dGraph, dToken, dNode, bKeep=False):
+        "generator: return matches where <dToken> “values” match <dNode> arcs"
         bTokenFound = False
         # token value
         if dToken["sValue"] in dNode:
-            if bDebug:
-                echo("  MATCH: " + dToken["sValue"])
-            yield { "iToken1": iToken1, "iNode": dNode[dToken["sValue"]] }
+            yield (" ", dToken["sValue"], dNode[dToken["sValue"]])
             bTokenFound = True
         if dToken["sValue"][0:2].istitle(): # we test only 2 first chars, to make valid words such as "Laissez-les", "Passe-partout".
             sValue = dToken["sValue"].lower()
             if sValue in dNode:
-                if bDebug:
-                    echo("  MATCH: " + sValue)
-                yield { "iToken1": iToken1, "iNode": dNode[sValue] }
+                yield (" ", sValue, dNode[sValue])
                 bTokenFound = True
         elif dToken["sValue"].isupper():
             sValue = dToken["sValue"].lower()
             if sValue in dNode:
-                if bDebug:
-                    echo("  MATCH: " + sValue)
-                yield { "iToken1": iToken1, "iNode": dNode[sValue] }
+                yield (" ", sValue, dNode[sValue])
                 bTokenFound = True
             sValue = dToken["sValue"].capitalize()
             if sValue in dNode:
-                if bDebug:
-                    echo("  MATCH: " + sValue)
-                yield { "iToken1": iToken1, "iNode": dNode[sValue] }
+                yield (" ", sValue, dNode[sValue])
                 bTokenFound = True
         # regex value arcs
         if dToken["sType"] not in frozenset(["INFO", "PUNC", "SIGN"]):
@@ -432,9 +428,7 @@ class TextParser:
                     if "¬" not in sRegex:
                         # no anti-pattern
                         if re.search(sRegex, dToken["sValue"]):
-                            if bDebug:
-                                echo("  MATCH: ~" + sRegex)
-                            yield { "iToken1": iToken1, "iNode": dNode["<re_value>"][sRegex] }
+                            yield ("~", sRegex, dNode["<re_value>"][sRegex])
                             bTokenFound = True
                     else:
                         # there is an anti-pattern
@@ -442,9 +436,7 @@ class TextParser:
                         if sNegPattern and re.search(sNegPattern, dToken["sValue"]):
                             continue
                         if not sPattern or re.search(sPattern, dToken["sValue"]):
-                            if bDebug:
-                                echo("  MATCH: ~" + sRegex)
-                            yield { "iToken1": iToken1, "iNode": dNode["<re_value>"][sRegex] }
+                            yield ("~", sRegex, dNode["<re_value>"][sRegex])
                             bTokenFound = True
         # analysable tokens
         if dToken["sType"][0:4] == "WORD":
@@ -452,9 +444,22 @@ class TextParser:
             if "<lemmas>" in dNode:
                 for sLemma in _oSpellChecker.getLemma(dToken["sValue"]):
                     if sLemma in dNode["<lemmas>"]:
-                        if bDebug:
-                            echo("  MATCH: >" + sLemma)
-                        yield { "iToken1": iToken1, "iNode": dNode["<lemmas>"][sLemma] }
+                        yield (">", sLemma, dNode["<lemmas>"][sLemma])
+                        bTokenFound = True
+            # phonetic similarity
+            if "<phonet>" in dNode:
+                for sPhonet in dNode["<phonet>"]:
+                    if sPhonet.endswith("!"):
+                        sPhon = sPhonet[0:-1]
+                        if dToken["sValue"] == sPhon:
+                            continue
+                        if dToken["sValue"][0:1].isupper():
+                            if dToken["sValue"].lower() == sPhon:
+                                continue
+                            if dToken["sValue"].isupper() and dToken["sValue"].capitalize() == sPhon:
+                                continue
+                    if phonet.isSimilAs(dToken["sValue"], sPhonet.rstrip("!")):
+                        yield ("#", sPhonet, dNode["<phonet>"][sPhonet])
                         bTokenFound = True
             # morph arcs
             if "<morph>" in dNode:
@@ -464,9 +469,7 @@ class TextParser:
                         if "¬" not in sSearch:
                             # no anti-pattern
                             if any(sSearch in sMorph  for sMorph in lMorph):
-                                if bDebug:
-                                    echo("  MATCH: $" + sSearch)
-                                yield { "iToken1": iToken1, "iNode": dNode["<morph>"][sSearch] }
+                                yield ("$", sSearch, dNode["<morph>"][sSearch])
                                 bTokenFound = True
                         else:
                             # there is an anti-pattern
@@ -475,17 +478,13 @@ class TextParser:
                                 # all morphologies must match with <sPattern>
                                 if sPattern:
                                     if all(sPattern in sMorph  for sMorph in lMorph):
-                                        if bDebug:
-                                            echo("  MATCH: $" + sSearch)
-                                        yield { "iToken1": iToken1, "iNode": dNode["<morph>"][sSearch] }
+                                        yield ("$", sSearch, dNode["<morph>"][sSearch])
                                         bTokenFound = True
                             else:
                                 if sNegPattern and any(sNegPattern in sMorph  for sMorph in lMorph):
                                     continue
                                 if not sPattern or any(sPattern in sMorph  for sMorph in lMorph):
-                                    if bDebug:
-                                        echo("  MATCH: $" + sSearch)
-                                    yield { "iToken1": iToken1, "iNode": dNode["<morph>"][sSearch] }
+                                    yield ("$", sSearch, dNode["<morph>"][sSearch])
                                     bTokenFound = True
             # regex morph arcs
             if "<re_morph>" in dNode:
@@ -495,9 +494,7 @@ class TextParser:
                         if "¬" not in sRegex:
                             # no anti-pattern
                             if any(re.search(sRegex, sMorph)  for sMorph in lMorph):
-                                if bDebug:
-                                    echo("  MATCH: @" + sRegex)
-                                yield { "iToken1": iToken1, "iNode": dNode["<re_morph>"][sRegex] }
+                                yield ("@", sRegex, dNode["<re_morph>"][sRegex])
                                 bTokenFound = True
                         else:
                             # there is an anti-pattern
@@ -506,63 +503,63 @@ class TextParser:
                                 # all morphologies must match with <sPattern>
                                 if sPattern:
                                     if all(re.search(sPattern, sMorph)  for sMorph in lMorph):
-                                        if bDebug:
-                                            echo("  MATCH: @" + sRegex)
-                                        yield { "iToken1": iToken1, "iNode": dNode["<re_morph>"][sRegex] }
+                                        yield ("@", sRegex, dNode["<re_morph>"][sRegex])
                                         bTokenFound = True
                             else:
                                 if sNegPattern and any(re.search(sNegPattern, sMorph)  for sMorph in lMorph):
                                     continue
                                 if not sPattern or any(re.search(sPattern, sMorph)  for sMorph in lMorph):
-                                    if bDebug:
-                                        echo("  MATCH: @" + sRegex)
-                                    yield { "iToken1": iToken1, "iNode": dNode["<re_morph>"][sRegex] }
+                                    yield ("@", sRegex, dNode["<re_morph>"][sRegex])
                                     bTokenFound = True
         # token tags
         if "aTags" in dToken and "<tags>" in dNode:
             for sTag in dToken["aTags"]:
                 if sTag in dNode["<tags>"]:
-                    if bDebug:
-                        echo("  MATCH: /" + sTag)
-                    yield { "iToken1": iToken1, "iNode": dNode["<tags>"][sTag] }
+                    yield ("/", sTag, dNode["<tags>"][sTag])
                     bTokenFound = True
         # meta arc (for token type)
         if "<meta>" in dNode:
             for sMeta in dNode["<meta>"]:
                 # no regex here, we just search if <dNode["sType"]> exists within <sMeta>
                 if sMeta == "*" or dToken["sType"] == sMeta:
-                    if bDebug:
-                        echo("  MATCH: *" + sMeta)
-                    yield { "iToken1": iToken1, "iNode": dNode["<meta>"][sMeta] }
+                    yield ("*", sMeta, dNode["<meta>"][sMeta])
                     bTokenFound = True
                 elif "¬" in sMeta:
                     if dToken["sType"] not in sMeta:
-                        if bDebug:
-                            echo("  MATCH: *" + sMeta)
-                        yield { "iToken1": iToken1, "iNode": dNode["<meta>"][sMeta] }
+                        yield ("*", sMeta, dNode["<meta>"][sMeta])
                         bTokenFound = True
-        if not bTokenFound and "bKeep" in dPointer:
-            yield dPointer
+        if not bTokenFound and bKeep:
+            yield (None, "", -1)
         # JUMP
-        # Warning! Recurssion!
+        # Warning! Recursion!
         if "<>" in dNode:
-            dPointer2 = { "iToken1": iToken1, "iNode": dNode["<>"], "bKeep": True }
-            yield from self._getNextPointers(dToken, dGraph, dPointer2, bDebug)
+            yield from self._getMatches(dGraph, dToken, dGraph[dNode["<>"]], bKeep=True)
 
     def parseGraph (self, dGraph, sCountry="FR", dOptions=None, bShowRuleId=False, bDebug=False, bContext=False):
         "parse graph with tokens from the text and execute actions encountered"
         lPointer = []
         bTagAndRewrite = False
-        for iToken, dToken in enumerate(self.lToken):
+        for iToken, dToken in enumerate(self.lTokens):
             if bDebug:
                 echo("TOKEN: " + dToken["sValue"])
             # check arcs for each existing pointer
             lNextPointer = []
             for dPointer in lPointer:
-                lNextPointer.extend(self._getNextPointers(dToken, dGraph, dPointer, bDebug))
+                for cActionType, sMatch, iNode in self._getMatches(dGraph, dToken, dGraph[dPointer["iNode"]]):
+                    if cActionType is None:
+                        lNextPointer.append(dPointer)
+                        continue
+                    if bDebug:
+                        echo("  MATCH: " + cActionType + sMatch)
+                    lNextPointer.append({ "iToken1": dPointer["iToken1"], "iNode": iNode })
             lPointer = lNextPointer
             # check arcs of first nodes
-            lPointer.extend(self._getNextPointers(dToken, dGraph, { "iToken1": iToken, "iNode": 0 }, bDebug))
+            for cActionType, sMatch, iNode in self._getMatches(dGraph, dToken, dGraph[0]):
+                if cActionType is None:
+                    continue
+                if bDebug:
+                    echo("  MATCH: " + cActionType + sMatch)
+                lPointer.append({ "iToken1": iToken, "iNode": iNode })
             # check if there is rules to check for each pointer
             for dPointer in lPointer:
                 #if bDebug:
@@ -591,19 +588,19 @@ class TextParser:
                     # TextProcessor [ option, condition, "~", replacement/suggestion/action, iTokenStart, iTokenEnd, bCaseSvty ]
                     # Disambiguator [ option, condition, "=", replacement/suggestion/action ]
                     # Tag           [ option, condition, "/", replacement/suggestion/action, iTokenStart, iTokenEnd ]
-                    # Immunity      [ option, condition, "!", "",                            iTokenStart, iTokenEnd ]
+                    # Immunity      [ option, condition, "!", option,                        iTokenStart, iTokenEnd ]
                     # Test          [ option, condition, ">", "" ]
                     if not sOption or dOptions.get(sOption, False):
-                        bCondMemo = not sFuncCond or getattr(gc_functions, sFuncCond)(self.lToken, nTokenOffset, nLastToken, sCountry, bCondMemo, self.dTags, self.sSentence, self.sSentence0)
+                        bCondMemo = not sFuncCond or getattr(gc_functions, sFuncCond)(self.lTokens, nTokenOffset, nLastToken, sCountry, bCondMemo, self.dTags, self.sSentence, self.sSentence0)
                         if bCondMemo:
                             if cActionType == "-":
                                 # grammar error
                                 iTokenStart, iTokenEnd, cStartLimit, cEndLimit, bCaseSvty, nPriority, sMessage, iURL = eAct
                                 nTokenErrorStart = nTokenOffset + iTokenStart  if iTokenStart > 0  else nLastToken + iTokenStart
-                                if "bImmune" not in self.lToken[nTokenErrorStart]:
+                                if "sImmunity" not in self.lTokens[nTokenErrorStart] or (self.lTokens[nTokenErrorStart]["sImmunity"] != "*" and sOption not in self.lTokens[nTokenErrorStart]["sImmunity"]):
                                     nTokenErrorEnd = nTokenOffset + iTokenEnd  if iTokenEnd > 0  else nLastToken + iTokenEnd
-                                    nErrorStart = self.nOffsetWithinParagraph + (self.lToken[nTokenErrorStart]["nStart"] if cStartLimit == "<"  else self.lToken[nTokenErrorStart]["nEnd"])
-                                    nErrorEnd = self.nOffsetWithinParagraph + (self.lToken[nTokenErrorEnd]["nEnd"] if cEndLimit == ">"  else self.lToken[nTokenErrorEnd]["nStart"])
+                                    nErrorStart = self.nOffsetWithinParagraph + (self.lTokens[nTokenErrorStart]["nStart"] if cStartLimit == "<"  else self.lTokens[nTokenErrorStart]["nEnd"])
+                                    nErrorEnd = self.nOffsetWithinParagraph + (self.lTokens[nTokenErrorEnd]["nEnd"] if cEndLimit == ">"  else self.lTokens[nTokenErrorEnd]["nStart"])
                                     if nErrorStart not in self.dError or nPriority > self.dErrorPriority.get(nErrorStart, -1):
                                         self.dError[nErrorStart] = self._createErrorFromTokens(sWhat, nTokenOffset, nLastToken, nTokenErrorStart, nErrorStart, nErrorEnd, sLineId, sRuleId, bCaseSvty, \
                                                                                                sMessage, _rules_graph.dURL.get(iURL, ""), bShowRuleId, sOption, bContext)
@@ -618,12 +615,12 @@ class TextParser:
                                 self._tagAndPrepareTokenForRewriting(sWhat, nTokenStart, nTokenEnd, nTokenOffset, nLastToken, eAct[2], bDebug)
                                 bChange = True
                                 if bDebug:
-                                    echo("    TEXT_PROCESSOR: [{}:{}]  > {}".format(self.lToken[nTokenStart]["sValue"], self.lToken[nTokenEnd]["sValue"], sWhat))
+                                    echo("    TEXT_PROCESSOR: [{}:{}]  > {}".format(self.lTokens[nTokenStart]["sValue"], self.lTokens[nTokenEnd]["sValue"], sWhat))
                             elif cActionType == "=":
                                 # disambiguation
-                                getattr(gc_functions, sWhat)(self.lToken, nTokenOffset, nLastToken)
+                                getattr(gc_functions, sWhat)(self.lTokens, nTokenOffset, nLastToken)
                                 if bDebug:
-                                    echo("    DISAMBIGUATOR: ({})  [{}:{}]".format(sWhat, self.lToken[nTokenOffset+1]["sValue"], self.lToken[nLastToken]["sValue"]))
+                                    echo("    DISAMBIGUATOR: ({})  [{}:{}]".format(sWhat, self.lTokens[nTokenOffset+1]["sValue"], self.lTokens[nLastToken]["sValue"]))
                             elif cActionType == ">":
                                 # we do nothing, this test is just a condition to apply all following actions
                                 if bDebug:
@@ -633,12 +630,12 @@ class TextParser:
                                 nTokenStart = nTokenOffset + eAct[0]  if eAct[0] > 0  else nLastToken + eAct[0]
                                 nTokenEnd = nTokenOffset + eAct[1]  if eAct[1] > 0  else nLastToken + eAct[1]
                                 for i in range(nTokenStart, nTokenEnd+1):
-                                    if "aTags" in self.lToken[i]:
-                                        self.lToken[i]["aTags"].update(sWhat.split("|"))
+                                    if "aTags" in self.lTokens[i]:
+                                        self.lTokens[i]["aTags"].update(sWhat.split("|"))
                                     else:
-                                        self.lToken[i]["aTags"] = set(sWhat.split("|"))
+                                        self.lTokens[i]["aTags"] = set(sWhat.split("|"))
                                 if bDebug:
-                                    echo("    TAG: {} >  [{}:{}]".format(sWhat, self.lToken[nTokenStart]["sValue"], self.lToken[nTokenEnd]["sValue"]))
+                                    echo("    TAG: {} >  [{}:{}]".format(sWhat, self.lTokens[nTokenStart]["sValue"], self.lTokens[nTokenEnd]["sValue"]))
                                 for sTag in sWhat.split("|"):
                                     if sTag not in self.dTags:
                                         self.dTags[sTag] = [nTokenStart, nTokenEnd]
@@ -651,15 +648,16 @@ class TextParser:
                                     echo("    IMMUNITY: " + sLineId + " / " + sRuleId)
                                 nTokenStart = nTokenOffset + eAct[0]  if eAct[0] > 0  else nLastToken + eAct[0]
                                 nTokenEnd = nTokenOffset + eAct[1]  if eAct[1] > 0  else nLastToken + eAct[1]
+                                sImmunity = sWhat or "*"
                                 if nTokenEnd - nTokenStart == 0:
-                                    self.lToken[nTokenStart]["bImmune"] = True
-                                    nErrorStart = self.nOffsetWithinParagraph + self.lToken[nTokenStart]["nStart"]
+                                    self.lTokens[nTokenStart]["sImmunity"] = sImmunity
+                                    nErrorStart = self.nOffsetWithinParagraph + self.lTokens[nTokenStart]["nStart"]
                                     if nErrorStart in self.dError:
                                         del self.dError[nErrorStart]
                                 else:
                                     for i in range(nTokenStart, nTokenEnd+1):
-                                        self.lToken[i]["bImmune"] = True
-                                        nErrorStart = self.nOffsetWithinParagraph + self.lToken[i]["nStart"]
+                                        self.lTokens[i]["sImmunity"] = sImmunity
+                                        nErrorStart = self.nOffsetWithinParagraph + self.lTokens[i]["nStart"]
                                         if nErrorStart in self.dError:
                                             del self.dError[nErrorStart]
                             else:
@@ -697,16 +695,16 @@ class TextParser:
     def _createErrorFromTokens (self, sSugg, nTokenOffset, nLastToken, iFirstToken, nStart, nEnd, sLineId, sRuleId, bCaseSvty, sMsg, sURL, bShowRuleId, sOption, bContext):
         # suggestions
         if sSugg[0:1] == "=":
-            sSugg = getattr(gc_functions, sSugg[1:])(self.lToken, nTokenOffset, nLastToken)
+            sSugg = getattr(gc_functions, sSugg[1:])(self.lTokens, nTokenOffset, nLastToken)
             lSugg = sSugg.split("|")  if sSugg  else []
         elif sSugg == "_":
             lSugg = []
         else:
             lSugg = self._expand(sSugg, nTokenOffset, nLastToken).split("|")
-        if bCaseSvty and lSugg and self.lToken[iFirstToken]["sValue"][0:1].isupper():
+        if bCaseSvty and lSugg and self.lTokens[iFirstToken]["sValue"][0:1].isupper():
             lSugg = list(map(lambda s: s.upper(), lSugg))  if self.sSentence[nStart:nEnd].isupper()  else list(map(lambda s: s[0:1].upper()+s[1:], lSugg))
         # Message
-        sMessage = getattr(gc_functions, sMsg[1:])(self.lToken, nTokenOffset, nLastToken)  if sMsg[0:1] == "="  else self._expand(sMsg, nTokenOffset, nLastToken)
+        sMessage = getattr(gc_functions, sMsg[1:])(self.lTokens, nTokenOffset, nLastToken)  if sMsg[0:1] == "="  else self._expand(sMsg, nTokenOffset, nLastToken)
         if bShowRuleId:
             sMessage += "  #" + sLineId + " / " + sRuleId
         #
@@ -755,9 +753,9 @@ class TextParser:
     def _expand (self, sText, nTokenOffset, nLastToken):
         for m in re.finditer(r"\\(-?[0-9]+)", sText):
             if m.group(1)[0:1] == "-":
-                sText = sText.replace(m.group(0), self.lToken[nLastToken+int(m.group(1))+1]["sValue"])
+                sText = sText.replace(m.group(0), self.lTokens[nLastToken+int(m.group(1))+1]["sValue"])
             else:
-                sText = sText.replace(m.group(0), self.lToken[nTokenOffset+int(m.group(1))]["sValue"])
+                sText = sText.replace(m.group(0), self.lTokens[nTokenOffset+int(m.group(1))]["sValue"])
         return sText
 
     def rewriteText (self, sText, sRepl, iGroup, m, bUppercase):
@@ -784,31 +782,35 @@ class TextParser:
         if sWhat == "*":
             # purge text
             if nTokenRewriteEnd - nTokenRewriteStart == 0:
-                self.lToken[nTokenRewriteStart]["bToRemove"] = True
+                self.lTokens[nTokenRewriteStart]["bToRemove"] = True
             else:
                 for i in range(nTokenRewriteStart, nTokenRewriteEnd+1):
-                    self.lToken[i]["bToRemove"] = True
+                    self.lTokens[i]["bToRemove"] = True
         elif sWhat == "␣":
             # merge tokens
-            self.lToken[nTokenRewriteStart]["nMergeUntil"] = nTokenRewriteEnd
+            self.lTokens[nTokenRewriteStart]["nMergeUntil"] = nTokenRewriteEnd
+        elif sWhat.startswith("␣"):
+            sWhat = self._expand(sWhat, nTokenOffset, nLastToken)
+            self.lTokens[nTokenRewriteStart]["nMergeUntil"] = nTokenRewriteEnd
+            self.lTokens[nTokenRewriteStart]["sMergedValue"] = sWhat[1:]
         elif sWhat == "_":
             # neutralized token
             if nTokenRewriteEnd - nTokenRewriteStart == 0:
-                self.lToken[nTokenRewriteStart]["sNewValue"] = "_"
+                self.lTokens[nTokenRewriteStart]["sNewValue"] = "_"
             else:
                 for i in range(nTokenRewriteStart, nTokenRewriteEnd+1):
-                    self.lToken[i]["sNewValue"] = "_"
+                    self.lTokens[i]["sNewValue"] = "_"
         else:
             if sWhat.startswith("="):
-                sWhat = getattr(gc_functions, sWhat[1:])(self.lToken, nTokenOffset, nLastToken)
+                sWhat = getattr(gc_functions, sWhat[1:])(self.lTokens, nTokenOffset, nLastToken)
             else:
                 sWhat = self._expand(sWhat, nTokenOffset, nLastToken)
-            bUppercase = bCaseSvty and self.lToken[nTokenRewriteStart]["sValue"][0:1].isupper()
+            bUppercase = bCaseSvty and self.lTokens[nTokenRewriteStart]["sValue"][0:1].isupper()
             if nTokenRewriteEnd - nTokenRewriteStart == 0:
                 # one token
                 if bUppercase:
                     sWhat = sWhat[0:1].upper() + sWhat[1:]
-                self.lToken[nTokenRewriteStart]["sNewValue"] = sWhat
+                self.lTokens[nTokenRewriteStart]["sNewValue"] = sWhat
             else:
                 # several tokens
                 lTokenValue = sWhat.split("|")
@@ -818,42 +820,51 @@ class TextParser:
                     return
                 for i, sValue in zip(range(nTokenRewriteStart, nTokenRewriteEnd+1), lTokenValue):
                     if not sValue or sValue == "*":
-                        self.lToken[i]["bToRemove"] = True
+                        self.lTokens[i]["bToRemove"] = True
                     else:
                         if bUppercase:
                             sValue = sValue[0:1].upper() + sValue[1:]
-                        self.lToken[i]["sNewValue"] = sValue
+                        self.lTokens[i]["sNewValue"] = sValue
 
     def rewriteFromTags (self, bDebug=False):
         "rewrite the sentence, modify tokens, purge the token list"
         if bDebug:
             echo("REWRITE")
-        lNewToken = []
+        lNewTokens = []
+        lNewTokens0 = []
         nMergeUntil = 0
         dTokenMerger = {}
-        for iToken, dToken in enumerate(self.lToken):
+        for iToken, dToken in enumerate(self.lTokens):
             bKeepToken = True
             if dToken["sType"] != "INFO":
                 if nMergeUntil and iToken <= nMergeUntil:
+                    # token to merge
                     dTokenMerger["sValue"] += " " * (dToken["nStart"] - dTokenMerger["nEnd"]) + dToken["sValue"]
                     dTokenMerger["nEnd"] = dToken["nEnd"]
                     if bDebug:
                         echo("  MERGED TOKEN: " + dTokenMerger["sValue"])
+                    dToken["bMerged"] = True
                     bKeepToken = False
+                    if iToken == nMergeUntil and "sMergedValue" in dTokenMerger:
+                        dTokenMerger["sValue"] = dTokenMerger["sMergedValue"]
+                        sSpaceFiller = " " * (dToken["nEnd"] - dTokenMerger["nStart"] - len(dTokenMerger["sMergedValue"]))
+                        self.sSentence = self.sSentence[:dTokenMerger["nStart"]] + dTokenMerger["sMergedValue"] + sSpaceFiller + self.sSentence[dToken["nEnd"]:]
                 if "nMergeUntil" in dToken:
-                    if iToken > nMergeUntil: # this token is not already merged with a previous token
+                    # first token to be merge with
+                    if iToken > nMergeUntil: # this token is not to be merged with a previous token
                         dTokenMerger = dToken
                     if dToken["nMergeUntil"] > nMergeUntil:
                         nMergeUntil = dToken["nMergeUntil"]
                     del dToken["nMergeUntil"]
                 elif "bToRemove" in dToken:
+                    # deletion required
                     if bDebug:
                         echo("  REMOVED: " + dToken["sValue"])
                     self.sSentence = self.sSentence[:dToken["nStart"]] + " " * (dToken["nEnd"] - dToken["nStart"]) + self.sSentence[dToken["nEnd"]:]
                     bKeepToken = False
             #
             if bKeepToken:
-                lNewToken.append(dToken)
+                lNewTokens.append(dToken)
                 if "sNewValue" in dToken:
                     # rewrite token and sentence
                     if bDebug:
@@ -866,5 +877,5 @@ class TextParser:
                     del dToken["sNewValue"]
         if bDebug:
             echo("  TEXT REWRITED: " + self.sSentence)
-        self.lToken.clear()
-        self.lToken = lNewToken
+        self.lTokens.clear()
+        self.lTokens = lNewTokens
