@@ -3,41 +3,43 @@ List of similar chars
 useful for suggestion mechanism
 """
 
-import re
-
 
 dDistanceBetweenChars = {
-    "a": {},
-    "e": {"é": 0.5},
-    "é": {"e": 0.5},
-    "i": {"y": 0.2},
-    "o": {},
-    "u": {},
-    "y": {"i": 0.3},
-    "b": {"d": 0.8, "h": 0.9},
-    "c": {"ç": 0.1, "k": 0.5, "q": 0.5, "s": 0.5, "x": 0.5, "z": 0.8},
-    "d": {"b": 0.8},
-    "f": {"v": 0.8},
-    "g": {"j": 0.5},
-    "h": {"b": 0.9},
-    "j": {"g": 0.5, "i": 0.9},
-    "k": {"c": 0.5, "q": 0.1, "x": 0.5},
-    "l": {"i": 0.9},
-    "m": {"n": 0.8},
-    "n": {"m": 0.8, "r": 0.9},
-    "p": {"q": 0.9},
-    "q": {"c": 0.5, "k": 0.1, "p": 0.9},
-    "r": {"n": 0.9, "j": 0.9},
-    "s": {"c": 0.5, "ç": 0.1, "x": 0.5, "z": 0.5},
-    "t": {"d": 0.9},
-    "v": {"f": 0.8, "w": 0.1},
-    "w": {"v": 0.1},
-    "x": {"c": 0.5, "k": 0.5, "q": 0.5, "s": 0.5},
-    "z": {"s": 0.5}
+    # dDistanceBetweenChars:
+    # - with Jaro-Winkler, values between 1 and 10
+    # - with Damerau-Levenshtein, values / 10 (between 0 and 1: 0.1, 0.2 ... 0.9)
+    #"a": {},
+    "e": {"é": 5},
+    "é": {"e": 5},
+    "i": {"y": 2},
+    #"o": {},
+    #"u": {},
+    "y": {"i": 3},
+    "b": {"d": 8, "h": 9},
+    "c": {"ç": 1, "k": 5, "q": 5, "s": 5, "x": 5, "z": 8},
+    "d": {"b": 8},
+    "f": {"v": 8},
+    "g": {"j": 5},
+    "h": {"b": 9},
+    "j": {"g": 5, "i": 9},
+    "k": {"c": 5, "q": 1, "x": 5},
+    "l": {"i": 9},
+    "m": {"n": 8},
+    "n": {"m": 8, "r": 9},
+    "p": {"q": 9},
+    "q": {"c": 5, "k": 1, "p": 9},
+    "r": {"n": 9, "j": 9},
+    "s": {"c": 5, "ç": 1, "x": 5, "z": 5},
+    "t": {"d": 9},
+    "v": {"f": 8, "w": 1},
+    "w": {"v": 1},
+    "x": {"c": 5, "k": 5, "q": 5, "s": 5},
+    "z": {"s": 5}
 }
 
 
 def distanceBetweenChars (c1, c2):
+    "returns a float between 0 and 1"
     if c1 == c2:
         return 0
     if c1 not in dDistanceBetweenChars:
@@ -318,8 +320,8 @@ d2toX = {
 dFinal1 = {
     "a": ("as", "at", "ant", "ah"),
     "A": ("AS", "AT", "ANT", "AH"),
-    "c": ("ch",),
-    "C": ("CH",),
+    "c": ("ch", "que"),
+    "C": ("CH", "QUE"),
     "e": ("et", "er", "ets", "ée", "ez", "ai", "ais", "ait", "ent", "eh"),
     "E": ("ET", "ER", "ETS", "ÉE", "EZ", "AI", "AIS", "AIT", "ENT", "EH"),
     "é": ("et", "er", "ets", "ée", "ez", "ai", "ais", "ait"),
@@ -334,6 +336,8 @@ dFinal1 = {
     "G": ("GH",),
     "i": ("is", "it", "ie", "in"),
     "I": ("IS", "IT", "IE", "IN"),
+    "k": ("que",),
+    "K": ("QUE",),
     "n": ("nt", "nd", "ns", "nh"),
     "N": ("NT", "ND", "NS", "NH"),
     "o": ("aut", "ot", "os"),
